#!/opt/imh-python/bin/python3
import os
import subprocess
import glob
import platform
import sys
from pathlib import Path

def run_cmd(command=None,check_flag=True):
    output = None
    if command and str(command):
        try:
            output = subprocess.run(
                command,
                shell=True,
                check = check_flag,
                stdout=subprocess.PIPE
                ).stdout.decode("utf-8").strip()
        except Exception as e:
            print(e)
    return output

def determine_server():
    """
    Determine Server Type for reporting
    """
    # Dedi / VPS
    get_kernel = """uname -r"""
    kernel = run_cmd(command=get_kernel)
    if 'vz7' in str(kernel):
        # VPS Types
        if os.path.exists('/var/cpanel'):
            return 'Managed_VPS'
        elif os.path.exists('/usr/local/cwp'):
            return 'CWP_VPS'
        elif os.path.exists('/etc/profile.d/wp3.sh'):
            return 'Platform_i'
        else:
            return 'cVPS'
    elif os.path.exists('/opt/dedrads'):
        # Dedicated Servers
        if os.path.exists('/var/cpanel'):
            return 'Managed_Dedi'
        else:
            return 'Bare_Metal_Dedi'
    # Shared Servers
    if 'ngx' in HOSTNAME:
        return 'Shared_NGINX'
    elif 'biz' in HOSTNAME:
        return 'Shared'
    elif 'res' in HOSTNAME:
        return 'Reseller'
    elif 'hub' in HOSTNAME:
        return 'Hub'


    return None

def determine_panel():
    """
    Determines type of Panel in use on server, i.e. cPanel, CWP, Platform i, etc
    """
    if os.path.exists('/var/cpanel'): # cPanel
        return 'cPanel'
    elif os.path.exists('/usr/local/cwp'): # CWP
        return 'CWP'
    elif os.path.exists('/etc/profile.d/wp3.sh'): # Platform i
        return 'Platform_i'


    return None


def determine_cms(docroot=None):
    """
    Determine CMS manually with provided document root by matching expected config files for known CMS
    """
    cms = None
    docroot = str(docroot)
    cms_dictionary = {
        f"""{docroot}/concrete.php""": 'Concrete',
        f"""{docroot}/Mage.php""": 'Magento',
        f"""{docroot}/configuration.php""": 'Joomla',
        f"""{docroot}/ut.php""": 'PHPList',
        f"""{docroot}/passenger_wsgi.py""": 'Django',
        f"""{docroot}/wp-login.php""": 'Wordpress'
        }
    if os.path.exists(f"""{docroot}/sites/default/settings.php"""):
        cms = 'Drupal'
    elif os.path.exists(f"""{docroot}/includes/configure.php"""):
        cms = 'Zen Cart'
    elif os.path.exists(f"""{docroot}/config/config.inc.php""") or os.path.exists(f"""{docroot}/config/settings.inc.php"""):
        cms = 'Prestashop'
    elif glob.glob(f"""{docroot}/*/passenger_wsgi.py"""):
        cms = 'Django'
    elif os.path.exists(f"""{docroot}/app/etc/env.php""") or os.path.exists(f"""{docroot}/app/etc/local.xml"""): # These are the config files for newer versions
        cms = 'Magento'
    elif os.path.exists(f"""{docroot}/vendor/laravel/"""):
        cms = 'Laravel'

    if not cms and docroot and docroot.split('/')[1] == 'home': # Ensure the directory string at least starts with /home
        document_root = Path(f"""{docroot}""")
        for f in document_root.glob('*'):
            if f.is_file():
                if str(f) in cms_dictionary:
                    cms = cms_dictionary[str(f)]
                elif f == f"""{docroot}/config.php""":
                    if os.path.exists(f"""{docroot}/admin/config.php"""):
                        cms = 'OpenCart'
                    else:
                        try:
                            with open(f, 'r', encoding='utf-8') as config:
                                first_line = str(config.readline())
                                if 'Moodle' in first_line:
                                    cms = 'Moodle'
                                else:
                                    cms = 'PhpBB'
                        except Exception as e:
                            print(f"""{f}: {e}""")
    return cms

def check_cpanel_subdomain(domain=None):
    """
    Check cPanel subdomains to determine if they're standalone subdomains or if they're for addon domains
    """
    if domain:
        with open('/etc/userdomains', 'r', encoding='utf-8') as userdomains:
            try:
                addon_domain = domain.split('.')[0] + '.' + domain.split('.')[2]
                for line in userdomains.readlines():
                    if addon_domain in line.split(':'):
                        return addon_domain
            except Exception as e:
                print(e)
    return domain

def identify_apache_config():
    """
    Function to determine Apache config file to use to check domains as different installs/versions can store the config in different files, starting w/ config for 2.2 and    checking different 2.4 config locations to end
    """
    config_file = None
    if os.path.exists('/usr/local/apache/conf/httpd.conf'):
        config_file = '/usr/local/apache/conf/httpd.conf'
    elif os.path.exists('/etc/httpd/conf/httpd.conf'):
        config_file = '/etc/httpd/conf/httpd.conf'
    elif os.path.exists('/etc/apache2/httpd.conf'):
        config_file = '/etc/apache2/httpd.conf'
    elif os.path.exists('/etc/apache2/conf/httpd.conf'):
        config_file = '/etc/apache2/conf/httpd.conf'
    return config_file

def doc_root_check(doc_root=None, domain=None, user=None):
    if doc_root and doc_root != '.' and doc_root != '/' and doc_root != './':
        bad_dirs = [
            f"""{doc_root}/.cache""",
            f"""{doc_root}/.well-known""",
            f"""{doc_root}/cgi-bin""",
            f"""{doc_root}/wp-admin""",
            f"""{doc_root}/wp-content""",
            f"""{doc_root}/cache""",
            f"""{doc_root}/temp""",
            f"""{doc_root}/tmp"""
            ]
        user_cms = determine_cms(docroot=doc_root)
        if user_cms:
            SITE_OWNERS.update({domain: user})
            CMS_INFO.update({domain: user_cms})
            old_total = CMS_COUNT.get(user_cms, 0)
            new_total = old_total + 1
            CMS_COUNT.update({user_cms: new_total})
            bad_dirs.append(doc_root)
        # Check sub directories
        dirs = Path(doc_root)
        for d in dirs.iterdir():
            if d.is_dir() and str(d) not in bad_dirs and str(d) != doc_root:
                dir_cms = determine_cms(docroot=str(d))
                if dir_cms:
                    limit = len(d.parts) - 1
                    sub_dir = d.parts[limit]
                    url = f"""{domain}/{sub_dir}"""
                    SITE_OWNERS.update({url: user})
                    old_total = CMS_COUNT.get(dir_cms, 0)
                    new_total = old_total + 1
                    CMS_COUNT.update({dir_cms: new_total})
                    CMS_INFO.update({url: dir_cms})
                    bad_dirs.append(str(d))
    return


def cms_counter_no_soft(users=[]):
    """
    Function to get counts of different CMS in use - without Softaculous - for VPS / Dedis
    """
    if CPANEL == 1: # Get doc roots from cPanel
        for user in users:
            bad_dirs = []
            domains_cmd = [f"""grep {user} /etc/userdomains | grep -Ev '^\*' | cut -d ':' -f 1"""]
            domains = run_cmd(domains_cmd)
            for domain in domains.split():
                user_cms = None
                if domain.count('.') > 1: # Determine if real sub-domain
                    new_domain = check_cpanel_subdomain(domain=domain)
                else:
                    new_domain = domain
                if NGINX == 1:
                    count = 0
                    limit = domain.count('.')
                    nginx_domain_string = ''
                    while count <= limit:
                        nginx_domain_string = nginx_domain_string + '_' + domain.split('.')[count]
                        count += 1
                    nginx_domain = user + '_' + nginx_domain_string + '.conf'
                    nginx_config = WEB_SERVER_CONFIG + nginx_domain
                    cmd = [f"""grep root {nginx_config} 2>/dev/null | """ + r"""awk '{print $2}' | uniq | tr -d ';'"""]
                else:
                    cmd = [f"""grep -E 'ServerName {domain}' {WEB_SERVER_CONFIG} -A3 | grep DocumentRoot | """ + r"""awk '{print $2}' | sort -g | uniq"""]
                doc_root = run_cmd(cmd)
                if doc_root:
                    doc_root_check(doc_root=str(doc_root), domain=new_domain, user=user)
    else: # No Softaculous / No cPanel
        if NGINX == 0: # Apache
            domains_cmd = [f"""grep ServerName '{WEB_SERVER_CONFIG}' | """ + r"""awk '{print $2}' | sort -g | uniq"""] # ServerAlias can also be used, but for the purposes of this seems most efficient to just check ServerName directives for domains
        else: # NGiNX
            domains_cmd = [f"""find '{WEB_SERVER_CONFIG}' -type f -name '*.conf' -print | xargs -d '\n' -l basename | cut -d '.' -f 1 | cut -d '_' -f 3- | tr '_' '.' | grep -v '^\*'"""]
        domains = run_cmd(domains_cmd)
        for domain in domains.split():
            if NGINX == 1:
                count = 0
                limit = domain.count('.')
                nginx_domain_string = ''
                while count <= limit:
                    nginx_domain_string = nginx_domain_string + '_' + domain.split('.')[count]
                    count += 1
                nginx_domain = user + '_' + nginx_domain_string + '.conf'
                nginx_config = WEB_SERVER_CONFIG + nginx_domain
                cmd = [f"""grep root {nginx_config} | """ + r"""awk '{print $2}' | uniq | tr -d ';'"""]
            else:
                cmd = [f"""grep 'ServerName {domain}' {WEB_SERVER_CONFIG} -A3 | grep DocumentRoot | """ + r"""awk '{print $2}' | uniq"""]
            doc_root = run_cmd(cmd)
            if doc_root:
                doc_root_check(doc_root=str(doc_root), domain=domain, user=user)
            else:
                print(f"""Unable to determine Document Root for {domain}\n""")
                continue
    return

def cms_counter(users=[]):
    """
    Function to get counts of different CMS in use using Softaculous - for Shared, Dedi, VPS
    """
    for user in users:
        try:
            cmd = f"""/usr/local/cpanel/3rdparty/bin/php /usr/local/cpanel/whostmgr/docroot/cgi/softaculous/cli.php --list_ins --user={user} | sed 1d | cut -d '|' -f 3,5"""
            try:
                user_installs = run_cmd(cmd)
            except Exception as e:
                print(e)
            if user_installs: 
                for install in user_installs.splitlines():
                    try:
                        user_cms = install.split('|')[0]
                        url = install.split('|')[1]
                        if 'www' in url.split('.'): # Strip 'www' from URLs
                            domain = url.split('.')[1] + '.' + url.split('.')[2]
                        else:
                            domain = url.split('/')[2]
                        if user_cms:
                            user_cms = user_cms.split()[0] # Remove version numbers which will cause duplicates in list, don't care if different versions
                        if user_cms and url and domain: 
                            cur_total = CMS_COUNT.get(user_cms, 0)
                            new_total = cur_total + 1
                            CMS_COUNT.update({user_cms: new_total}) # count the totals of different CMS in use, will output at the end
                            CMS_INFO.update({url: user_cms})
                            SITE_OWNERS.update({url: user})
                    except:
                        continue
            else:
                try:
                    cms_counter_no_soft(users=[user]) # Since softaculous was found but no installs identified, check for installs without Softaculous instead - pass user as single item list as function expects list of users
                except Exception as e:
                    print(e)
        except Exception as e:
            print(e)
            continue
    return

def main():
    """
    Main function, initializes global variables as globals, gets hostname and call relevant checks after determining enviroment
    """

    # Global Variables
    global CPANEL
    global NGINX
    global WEB_SERVER_CONFIG
    global CMS_COUNT
    global CMS_INFO
    global CHECK_NGINX_STATUS
    global CHECK_APACHE_STATUS
    global HOSTNAME
    global SITE_OWNERS
    try:
        HOSTNAME = platform.node().split('.')[0]
    except:
        sys.exit('Unable to determine hostname!\n')


    # Determine Web Server
    apache_status = ["""service httpd status &>/dev/null;echo $?"""]
    nginx_status = ["""service nginx status &>/dev/null;echo $?"""]
    check_apache_status = None
    check_nginx_status = None
    try:
        check_apache_status = run_cmd(apache_status)
        check_nginx_status = run_cmd(nginx_status)
    except:
        sys.exit('Unable to determine Web Server!\nExiting...\n')
    if check_nginx_status == '0': # NGiNX is in use
        NGINX = 1
        WEB_SERVER_CONFIG = '/etc/nginx/vhosts/'
    elif check_apache_status == '0': # Apache is in use
        WEB_SERVER_CONFIG = identify_apache_config()
        if not WEB_SERVER_CONFIG:
            sys.exit('Unable to determine Web Server Config!\nExiting...\n')
    else:
        sys.exit('Unable to determine Web Server Config!\nExiting...\n')

    # Enumerate Users
    user_list = []
    if os.path.exists('/var/cpanel/users'):
        CPANEL = 1
        users_dir = Path('/var/cpanel/users')
        for u in users_dir.iterdir(): 
            limit = len(u.parts) - 1
            if not u.is_dir() and u.parts[limit] != 'system' and '.bak' not in u.parts[limit]:
                user_list.append(u.parts[limit])
    else:
        users_dir = Path('/home')
        try:
            with open('/etc/passwd') as passwd:
                passwd_file = passwd.readlines()
        except:
            sys.exit('Unable to read /etc/passwd!\nExiting...')
        for u in users_dir.iterdir():
            if u.is_dir():
                limit = len(u.parts) - 1
                for line in passwd_file:
                    if u.parts[limit] == line.split(':')[0]:
                        user_list.append(u.parts[limit])
    if user_list == []:
        sys.exit('User List empty!\nExiting...')

    # Check for Softaculous and run appropriate counter function
    if os.path.exists('/var/softaculous'):
        cms_counter(users=user_list)
    else:
        cms_counter_no_soft(users=user_list)

    # Output
    try: # Verbose
        if sys.argv[1] and sys.argv[1] == '-v' or sys.argv[1] == '--verbose':
            st = determine_server()
            pt = determine_panel()
            for url,cms in CMS_INFO.items():
                site_owner = SITE_OWNERS.get(url, None)
                print(f"""{st} {pt} {site_owner} {url} {cms}""")
    except: # Not Verbose
        for url,cms in CMS_INFO.items():
            site_owner = SITE_OWNERS.get(url, None)
            print(f"""{site_owner} {url} {cms}""")


# Global Variables
CPANEL = 0
NGINX = 0
WEB_SERVER_CONFIG = None
CMS_COUNT = {} # Counts the number of installs by CMS for output at the end
CMS_INFO = {} # Stores the URL and CMS info for each site
SITE_OWNERS = {}
CHECK_NGINX_STATUS = None
CHECK_APACHE_STATUS = None
HOSTNAME = None

if __name__ == "__main__":
    main()
